import java.io.*;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.HashSet;

public final class Forest {           
  public HashSet<Node> forestTreeRoots;
  public HashMap<Integer, Routine> routines;
  public Tool tool;
  public int chosenK;
  public int bufsize;
  public int numNodes;
  public int numRoutines;   
  public long blppHashEvents;
  public long kblppHashEvents;
        
  public Forest(Tool tool, int chosenK, int bufsize) {
    this.forestTreeRoots=new HashSet<Node>();
    this.routines=new HashMap<Integer, Routine>();
    this.tool=tool;
    this.chosenK=chosenK;
    this.bufsize=bufsize;
    this.numNodes=0;
    this.numRoutines=0;    
    this.blppHashEvents=0;
    this.kblppHashEvents=0;
  }
  
  public final void clean() {
    // do not free nodes inside forestTreeRoots...
    forestTreeRoots.clear();
    forestTreeRoots=null;

    // ...since they will be freed now
    for (Routine r : routines.values())
      r.clean();
    routines.clear();
    routines=null;    
  }    
  
  public final void performCounterAnalysis(ProcessedLogStats stats) {
    long globalSumOfBLPPcounters = 0;
    long globalSumOfKBLPPcounters = 0;    
    int routinesWithCycles = 0;
      
    for (Routine r : routines.values()) {
      if (r.containsCycles) routinesWithCycles++;
      r.performCounterAnalysis(stats);      
      globalSumOfBLPPcounters += r.routineStats.sumOfBLPPcounters;  
      globalSumOfKBLPPcounters += r.routineStats.sumOfKBLPPcounters;
    }
              
    // write stats
    stats.sumOfBLPPcounters=globalSumOfBLPPcounters;
    stats.sumOfKBLPPcounters=globalSumOfKBLPPcounters;
    stats.processedEvents=globalSumOfBLPPcounters+globalSumOfKBLPPcounters;
    if (stats.numKSFinternalNodes>0)
      stats.avgKSFdegree=((double)stats.numKSFchildren)/stats.numKSFinternalNodes;
    if (stats.numKSFchildren>0)
      stats.avgKSFperChildLookup=((double)stats.sumKSFchildLookups)/stats.sumKSFchildCounters;
    
    String s;
    double fraction = 0;    
    DecimalFormat myFormat = new DecimalFormat("#0.##");
    
    if (tool==Tool.BLPP || tool==Tool.TBLPP) {      
      System.out.println("Sum of BLPP counters: "+globalSumOfBLPPcounters);      
      System.out.println("BLPP hash accesses: "+(globalSumOfBLPPcounters+numNodes));
      System.out.println("Hash get operations: "+globalSumOfBLPPcounters);
      System.out.println("Hash put operations: "+numNodes);    
      // write stats
      stats.numBLPPGetEvents=globalSumOfBLPPcounters;
      stats.numBLPPPutEvents=numNodes;
    } else {
      if (numRoutines>0)
        fraction=((double) routinesWithCycles)/numRoutines;
      s = myFormat.format(fraction);
      System.out.println("Routines with cycles: "+routinesWithCycles);    
      System.out.println("Fraction of routines with cycles: "+s);      
      System.out.println("Sum of k-SF counters for out-of-cycle paths: "+globalSumOfBLPPcounters);
      System.out.println("Sum of k-SF counters for k-iteration paths: "+globalSumOfKBLPPcounters);          
      System.out.println("k-SF nodes for out-of-cycle paths: "+stats.numBLPPNodes);
      System.out.println("k-SF nodes for k-iteration paths: "+stats.numKBLPPNodes);
      
      fraction = 0;
      if (globalSumOfBLPPcounters > 0 || globalSumOfKBLPPcounters>0)
        fraction = ((double) globalSumOfKBLPPcounters)/(globalSumOfBLPPcounters+globalSumOfKBLPPcounters);    
      s = myFormat.format(fraction);
      System.out.println("Fraction of total counter sum from k-iteration paths: "+s);  
      
      // reconstruct number of get and put operations
      stats.numBLPPGetEvents=globalSumOfBLPPcounters;
      stats.numBLPPPutEvents=stats.numBLPPNodes;
      stats.numKBLPPGetEvents=kblppHashEvents-stats.numKBLPPTrees;
      stats.numKBLPPPutEvents=stats.numKBLPPTrees;

      System.out.println("Hash accesses for out-of-cycle paths: "+blppHashEvents);        
      System.out.println("Hash get events for out-of-cycle paths: "+stats.numBLPPGetEvents);
      System.out.println("Hash put events for out-of-cycle paths: "+stats.numBLPPPutEvents);
        
      System.out.println("Hash accesses for k-iteration paths: "+kblppHashEvents);
      System.out.println("Hash get events for k-iteration paths: "+stats.numKBLPPGetEvents);
      System.out.println("Hash put events for k-iteration paths: : "+stats.numKBLPPPutEvents);
      
      fraction = stats.computeFractionOfAccessesForCyclesWrtBLPP();
      s=myFormat.format(fraction);
      System.out.println("Fraction of hash accesses made for k-iteration paths wrt traditional BLPP: "+s);
      
      fraction = stats.computeFractionOfAccessesWrtBLPP();
      s=myFormat.format(fraction);
      System.out.println("Fraction of total hash accesses wrt traditional BLPP: "+s);
          
      if (tool==Tool.SKBLPP) {
        // perform avg iterations analysis for each routine
        for (Routine r : routines.values()) {
          if (!r.containsCycles) continue;
          performAvgIterationsAnalysis(r.routineStats, stats);
        }
        // compute statistics for the whole thread        
        for (int i=0; i<stats.valuesForK.length; i++) {
          if (stats.valuesForK[i] > chosenK) break;
          if (globalSumOfKBLPPcounters>0)
            fraction = ((double) stats.eventsWithinK[i])/(globalSumOfKBLPPcounters);
          else fraction=0;  
          s = myFormat.format(fraction);
          System.out.println("Fraction of KBLPP events inside loops of length within "+(stats.valuesForK[i]+1)+": "+s);
        }    
      }  
    }         
    System.out.println("Number of processed events: "+stats.processedEvents);
    
    System.out.println("Number of k-SF internal nodes: "+stats.numKSFinternalNodes);
    s = myFormat.format(stats.avgKSFdegree);
    System.out.println("Avg k-SF node degree: "+s);
    s = myFormat.format(stats.avgKSFperChildLookup);
    System.out.println("Avg k-SF child lookups: "+s);
  }
  
  private final static void performAvgIterationsAnalysis(RoutineStats routineStats, ProcessedLogStats stats) {    
    for (int i=0; i<stats.valuesForK.length; i++) {
      if (routineStats.avgIterations <= (stats.valuesForK[i]+1))
        stats.eventsWithinK[i]+=routineStats.sumOfKBLPPcounters;
    }
  }     
    
  public final void dumpKSLAB(String filename, boolean generateSVG) {
    String threadName;
    if (filename.length()<=4) {
    System.out.println("Please do not rename logfiles!");
      threadName = filename+"-out"; 
    } else {   
      threadName = filename.substring(0, filename.length()-4);
    }        
    String destDir = threadName+"/kslab-"+chosenK;
    System.out.println("Creating output files in "+destDir+"/ folder...");
    
    File dir = new File(destDir+"/dot");
    dir.mkdirs();
    
    Runtime rt = null;    
    if (generateSVG) {
      dir = new File(destDir+"/svg");
      dir.mkdirs();
      rt = Runtime.getRuntime();
    }  
    
    for (Routine r : routines.values()) {
      if (!r.containsCycles) continue;
      String dotname = destDir + "/dot/" + r.cmid + ".dot";
      PrintStream ps = null;
      try {
        ps = new PrintStream(new FileOutputStream(dotname));
      } catch (IOException ex) {
        System.out.println("Error: cannot open output file "+dotname+ " - exiting...");
        System.out.println(ex);
        System.exit(1);
      }    
      r.dumpKSLAB(ps, this);
      ps.close();
        
      if (generateSVG) {
        String svgname = destDir + "/svg/" + r.cmid + ".svg";
        String command = "/usr/bin/dot -Tsvg "+dotname+" -o "+svgname;      
        try {      
          Process pr = rt.exec(command);
          //~ pr.waitfor();
        } catch (IOException e) {
          System.out.println("Warning: unable to generate SVG graph "+svgname);
          System.out.println(e);
        }        
      }
    }  
  }
  
  }
  
