import java.awt.Color;
import java.io.*;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Scanner;
import java.util.Stack;
import java.util.StringTokenizer;

public final class ProcessLog {

  // configuration parameters to be set by user
  public static boolean dumpKSFs = false;
  public static boolean dumpKIPFs = false;
  public static boolean pruneKIPFs = false;
  public static float rootThreshold = 0;
  public static float treeRootThreshold = 0;
  public static float childThreshold = 0;

  // useful to show errors for file processing methods
  private static String currentFilename = null;
  private static int currentRow = -1;
  private static int lastK = -1;
  private static Tool lastTool = null;
    
  // automatically reconstruct routine DAG
  public static boolean reconstructRoutineGraph = false;
  public static String routineDirName = null;
  
  private static HashMap<Integer, RoutineGraph> graphMap = new HashMap<Integer, RoutineGraph>();
  

  // auxiliary method to print errors
  private static final void brokenLogfile(String error) {    
    if (error!=null) {
      if (ProcessLog.currentRow>0) System.out.println("ERROR (row "+ProcessLog.currentRow+"): "+error); 
      else System.out.println("ERROR: "+error);
    }  
    System.out.println("Broken file \""+currentFilename+"\" - exiting...");
    System.exit(1);
  }

  private static final RoutineGraph getRoutineGraph(int cmid) {
    Integer i = Integer.valueOf(cmid);
    RoutineGraph g = graphMap.get(i);
    if (g==null) {      
      ProcessedRoutine p = new ProcessedRoutine(cmid);      
      g = p.process();
      if (g!=null) {
        //~ System.out.println("Routine "+cmid+" processed!");
        graphMap.put(i, g);        
      } else {
        System.out.println("WARNING: routine "+cmid+" not found!"); // TODO
      }              
    }
    return g;  
  }  
  
  public static final void dumpGraphvizAuxBLPP(ArrayList<Node> list, PrintStream ps) {
    for (Iterator<Node> itr = list.iterator(); itr.hasNext(); ) {
      Node node = itr.next();              

      if (node.isSpecial || node.type!=NodeType.BLPP)
        System.out.println("WARNING: BLPP node ("+node.cmid+", "+node.pathNum+") is broken!");
        
      String nodeInfo="\""+node.nodeID+"\" " + "[label=\""+ node.pathNum + " ("+node.count + ")"+
                      "\", "+"shape=ellipse, style=filled, fillcolor=blue];";        
      
      ps.println(nodeInfo);
    }      
  }    
  
  // print a set of BLPP (isolated) nodes as a table
  public static final void dumpGraphvizAuxBLPPAsTable(ArrayList<Node> list, PrintStream ps) {
    
    if (list==null || ps==null) return;
    
    ps.println("Out-of-cycle-paths [shape=none, margin=0, label=<");
    ps.println("<TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"2px\" CELLPADDING=\"2px\">");
    ps.println("<TH><TD>Acyclic paths</TD></TH>");
    ps.println("<TR><TD>Path Num</TD><TD>Count</TD></TR>");
    
    Collections.sort(list, new Comparator<Node>() {
      public int compare(Node a, Node b) {
        Long l1 = new Long(a.count);
        Long l2 = new Long(b.count);
        return l2.compareTo(l1); // sort by decreasing frequencies        
      }    
    });
        
    for (Iterator<Node> itr = list.iterator(); itr.hasNext(); ) {
      Node node = itr.next();
      ps.println("<TR><TD>"+node.pathNum+"</TD><TD>"+node.count+"</TD></TR>");      
      //~ String nodeInfo="\""+node.nodeID+"\" " + "[label=\""+ node.pathNum + " ("+node.count + ")"+ "\", ";
              //~ 
      //~ if (node.isSpecial) nodeInfo += "shape=box";
      //~ else nodeInfo += "shape=ellipse";
      //~ nodeInfo+=", style=filled, fillcolor=blue];";
             //~ 
      //~ ps.println(nodeInfo);                
    }        
    
    ps.println("</TABLE>");
    ps.println(">];");    
  }
  
  public static final void dumpGraphvizAux(Node parent, Node node, PrintStream ps) {
    if (node==null || ps==null) return;                
        
    if (parent!=null) {
      // edge: parent -> node
      ps.println("\""+parent.nodeID+"\" -> \""+node.nodeID+"\";");      
    }   
        
    String nodeInfo="\""+node.nodeID+"\" " + "[label=\""+ node.pathNum + " ("+node.count + ")"+ "\", ";
              
    if (node.isSpecial) {
      nodeInfo += "shape=box, style=filled";
      switch (node.type) {
        case BEGIN: nodeInfo += ", fillcolor=green];"; break;
        case EXIT: nodeInfo += ", fillcolor=red];"; break;
        case BLPP: nodeInfo += ", fillcolor=blue];"; // TODO might never happen
                   System.out.println("Warning: BLPP node ("+node.cmid+", "+node.pathNum+") in abnormal position!");
                   break;
        default: nodeInfo += ", fillcolor=orange];"; break;
      }    
    } else {
      nodeInfo += "shape=ellipse";
      switch (node.type) {
        case BEGIN: nodeInfo += ", style=filled, fillcolor=green];"; break;
        case EXIT: nodeInfo += ", style=filled, fillcolor=red];"; break;
        case BLPP: nodeInfo += ", style=filled, fillcolor=blue];"; break;
        default: nodeInfo += "];"; break;
      }
    }
        
    ps.println(nodeInfo);
                
    // process children recursively...
    for (Node tmp=node.firstChild; tmp!=null; tmp=tmp.nextSibling)
      dumpGraphvizAux(node, tmp, ps);
  }
  
  private static final String getColor(double percentage) {
    Color color1 = Color.WHITE;
    Color color2 = Color.GRAY;
    
    double ratio = percentage/100.0;
    
    int red = (int) (color2.getRed() * ratio + color1.getRed() * (1 - ratio));
    int green = (int) (color2.getGreen() * ratio + color1.getGreen() * (1 - ratio));
    int blue = (int) (color2.getBlue() * ratio + color1.getBlue() * (1 - ratio));
    
    String s = "#";
    s+= Integer.toHexString(red);
    s+= Integer.toHexString(green);
    s+= Integer.toHexString(blue);
    return s;
  }    
  
  // print a tree as a subgraph (you should manually add "subgraph sX {" and "}" to ps)
  public static final void dumpGraphvizAuxWithPruning(Node parent, Node node, PrintStream ps, Node treeRoot) {
    if (node==null || ps==null) return;                
    
    float ratio1 = 100;
    float ratio2 = 100;
        
    if (parent!=null) {
      // edge: parent -> node
      
      //~ ps.println("\""+parent.nodeID+"\" -> \""+node.nodeID+"\";");      
      
      ratio1 = (float)node.count/parent.count*100;
      ratio2 = (float)node.count/treeRoot.count*100;
      ps.println("\""+parent.nodeID+"\" -> \""+node.nodeID+"\" [ label= \" "+String.format("%.1f", ratio1)+"% ("+String.format("%.1f", ratio2)+"%)\" ];");      
      
    }   
        
    String nodeInfo="\""+node.nodeID+"\" " + "[label=\""+ node.pathNum + " ("+node.count + ")"+ "\", ";
              
    if (node.isSpecial) {
      nodeInfo += "shape=box, style=filled";
      switch (node.type) {
        case BEGIN: nodeInfo += ", fillcolor=green];"; break;
        case EXIT: nodeInfo += ", fillcolor=red];"; break;
        case BLPP: nodeInfo += ", fillcolor=blue];"; // TODO might never happen
                   System.out.println("Warning: BLPP node ("+node.cmid+", "+node.pathNum+") in abnormal position!");
                   break;
        default: nodeInfo += ", fillcolor=orange];"; break;
      }    
    } else {
      nodeInfo += "shape=ellipse";
      switch (node.type) {
        case BEGIN: nodeInfo += ", style=filled, fillcolor=green];"; break;
        case EXIT: nodeInfo += ", style=filled, fillcolor=red];"; break;
        case BLPP: nodeInfo += ", style=filled, fillcolor=blue];"; break;
        default: nodeInfo += ", style=filled, fillcolor=\""+getColor(ratio2)+"\"];"; break;
      }
    }
        
    ps.println(nodeInfo);
                
    // process children recursively...
    for (Node tmp=node.firstChild; tmp!=null; tmp=tmp.nextSibling) {
      if (tmp.count < node.count*ProcessLog.childThreshold || tmp.count < treeRoot.count*ProcessLog.treeRootThreshold)
        continue; // second pruning stage
      dumpGraphvizAuxWithPruning(node, tmp, ps, treeRoot);    
    }    
  }
   
  private static final int parseInt(String s) {
    return parseInt(s, 10);    
  }
  
  private static final int parseInt(String s, int base) {      
    int ret = 0;
    try {
      ret = Integer.parseInt(s, base);        
    } catch (NumberFormatException e) {
      brokenLogfile("numerical field not found");
      System.out.println(e);
    }
    return ret;
  }  
  
  private static final long parseLong(String s) {
    long ret = 0;
    try {
      ret = Long.parseLong(s);        
    } catch (NumberFormatException e) {
      brokenLogfile("numerical field not found");
      System.out.println(e);
    }
    return ret;    
  }    
  
  private static final float parseFloat(String s) {
    float ret = 0;
    try {
      ret = Float.parseFloat(s);
    } catch (NumberFormatException e) {
      brokenLogfile("numerical field not found");
      System.out.println(e);
    }
    return ret;      
  }
  
  public static final class ProcessedRoutine {
        
    public int cmid;
    
    public String tmpFilename;
    public int tmpCurrentRow;    
    
    public ProcessedRoutine(int cmid) {
      this.cmid=cmid;      
    }
    
    private static final int parseIntForEdge(String s) {      
      int ret = 0;
      try {
        ret = Integer.parseInt(s);        
      } catch (NumberFormatException e) {
        reportEdgeError("numerical field not found in routine edge row");
        System.out.println(e);
      }
      return ret;
    }
    
    private static final float parseFloatForEdge(String s) {
      float ret = 0;
      try {
        ret = Float.parseFloat(s);
      } catch (NumberFormatException e) {
        reportEdgeError("numerical field not found in routine edge row");
        System.out.println(e);
      }
      return ret;    
    }    
    
    private static final void reportEdgeError(String message) {
      if (message!=null)
        System.out.println("ERROR: "+message+" - exiting...");
      else
        System.out.println("ERROR: unknown error in a routine edge row - exiting...");  
      System.exit(1);
    }    
        
    public final RoutineGraph process() {
      // save global variables
      tmpFilename=ProcessLog.currentFilename;
      tmpCurrentRow=ProcessLog.currentRow;
      
      //~ RoutineGraph graph = new 
      String filename = ProcessLog.routineDirName+"/"+cmid+".info";
      File f = new File(filename);
      if (!f.exists()) return null;
      
      ProcessLog.currentFilename=filename;
      ProcessLog.currentRow=0;
      
      // process header
      
      RoutineGraph graph = null;
          
      try {
        // open file as a stream
        FileInputStream fstream = new FileInputStream(filename);        
        DataInputStream in = new DataInputStream(fstream);
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
                
        String s1, s2;
        StringTokenizer st;
        
        // process header
        String headerBroken = "a routine info header row is broken";
        String headerPrematureEnd = "routine info header ends prematurely";
        String headerUnexpectedFields = "a routine info header row contains unexpected field(s)";
        int cmid, optLevel;
        String methodName, shortClassName, pkgName, srcFile;        
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken);
        if (st.nextToken().equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);  // TODO && anche altrove
        cmid = parseInt(st.nextToken());
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken);
        if (st.nextToken().equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);  // TODO
        methodName = st.nextToken();
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken);
        if (st.nextToken().equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);  // TODO
        shortClassName = st.nextToken();
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken);
        //~ s2 = st.nextToken();
        // TODO - workaround for pkgname
        //~ if (s2.equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);        
        //~ pkgName = st.nextToken();
        if (!st.nextToken().equals("c")) brokenLogfile(headerBroken);
        if (st.hasMoreTokens())        
          pkgName = st.nextToken();
        else
          pkgName = "N/A";  
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken);
        if (st.nextToken().equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);    
        srcFile = st.nextToken();
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);
        
        s1 = br.readLine();
        if (s1==null) brokenLogfile(headerPrematureEnd);
        ProcessLog.currentRow++;
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile(headerBroken); // TODO
        if (st.nextToken().equals("c") && !st.hasMoreTokens()) brokenLogfile(headerBroken);
        optLevel = parseInt(st.nextToken());
        if (st.hasMoreTokens()) brokenLogfile(headerUnexpectedFields);

        graph = new RoutineGraph(cmid, methodName, shortClassName, pkgName, srcFile, optLevel);
        
        // process nodes and edges
        boolean graphEnded = false;
        Stack<String> edges = new Stack<String>();
        HashMap<String, RoutineNode> nodes = graph.nodeMap;
        String nodeRowBroken = "a routine node row is broken";        
                
        while ((s1 = br.readLine())!=null) {
          ProcessLog.currentRow++;
          st = new StringTokenizer(s1, " ");
          if (!st.hasMoreTokens()) brokenLogfile("a routine graph row is broken");
          s2 = st.nextToken();
          if (s2.equals("p")) {
            graphEnded = true;
            break;  
          }
          if (s2.equals("a")) {
            edges.push(s1);
            continue;
          }    
          if (!s2.equals("v")) brokenLogfile("unexpected row in file");
          
          // fields for new RoutineNode
          String ID, nextUnprofiled;
          int numPaths;
          RoutineNodeType type;
                    
          if (!st.hasMoreTokens()) brokenLogfile(nodeRowBroken);
          ID=st.nextToken();
          
          if (!st.hasMoreTokens()) brokenLogfile(nodeRowBroken);
          nextUnprofiled=st.nextToken();
          
          if (!st.hasMoreTokens()) brokenLogfile(nodeRowBroken);
          numPaths=parseInt(st.nextToken());
          
          if (!st.hasMoreTokens()) brokenLogfile(nodeRowBroken);
          type = RoutineNodeType.getTypeForString(st.nextToken());
          if (type==null) brokenLogfile("invalid routine node type");
          
          if (st.hasMoreTokens()) brokenLogfile("unexpected field(s) in a routine node row");
          
          RoutineNode node = new RoutineNode(ID, nextUnprofiled, numPaths, type);
          if (type==RoutineNodeType.BEGIN) graph.begin=node;
          else if (type==RoutineNodeType.END) graph.end=node;
          
          nodes.put(ID, node);                                                    
       }
       
       if (!graphEnded) brokenLogfile("premature end of file");       
       if (graph.begin==null) brokenLogfile("missing begin node in routine graph");
       if (graph.end==null) brokenLogfile("missing end node in routine graph");
       
       if (!st.hasMoreTokens()) brokenLogfile("routine graph info row is broken");
       
       // get numNodes and numEdges
       int numNodes = parseInt(st.nextToken());
       if (!st.hasMoreTokens()) brokenLogfile("routine graph info row is broken");
       int numEdges = parseInt(st.nextToken());
       
       if (numNodes != nodes.size()) {
         System.out.println("ERROR: wrong number of nodes in routine graph");
         System.exit(1);    
       }
       
       if (numEdges != edges.size()) {
         System.out.println("ERROR: wrong number of nodes in routine graph");
         System.exit(1);    
       }
       
       if (st.hasMoreTokens()) brokenLogfile("unexpected field(s) in routine graph info row");               
       
       // create edges and attach them to nodes
       String unknownNode = "a routine edge row refers to a nonexistent node";
       String brokenEdgeRow = "a routine edge row is broken";
       while (!edges.isEmpty()) {
         s1 = edges.pop();         
         
         RoutineNode src, dest;
         int cycleIndex, index, origMethodID, val;
         float prob;
         boolean isDummy=false;
         
         st = new StringTokenizer(s1, " ");         
         st.nextToken(); // I already know that the first token is "a"
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         src=nodes.get(st.nextToken());
         if (src==null) reportEdgeError(unknownNode);
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         dest=nodes.get(st.nextToken());
         if (dest==null) reportEdgeError(unknownNode);
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         val=parseIntForEdge(st.nextToken());
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         index=parseIntForEdge(st.nextToken());
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         prob=parseFloatForEdge(st.nextToken());
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         if (parseIntForEdge(st.nextToken())>0) isDummy=true;
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         origMethodID=parseIntForEdge(st.nextToken());
         
         if (!st.hasMoreTokens()) reportEdgeError(brokenEdgeRow);
         cycleIndex=parseIntForEdge(st.nextToken());
         
         if (st.hasMoreTokens()) reportEdgeError("unexpected field(s) in a routine edge row");
         
         //~ public RoutineEdge(RoutineNode src, RoutineNode dest, int val, int index, float prob, boolean isDummy, int origMethodID, int cycleIndex) {
         RoutineEdge edge = new RoutineEdge(src, dest, val, index, prob, isDummy, origMethodID, cycleIndex);
         src.outEdges.add(edge);
         //~ dest.inEdges.add(edge);         
       }    
              
       // restore ProcessLog.currentFilename and ProcessLog.currentRow
       ProcessLog.currentFilename=tmpFilename;
       ProcessLog.currentRow=tmpCurrentRow;
       
       // close file
       in.close();
       fstream.close();          
      
      } catch (Exception e) {
        System.out.println("Unknown error when processing file \""+filename+"\" - exiting...");
        System.out.println(e);
        System.exit(1);
      }
                  
      return graph;
    }       
      
      
  }    
  
  public static final class ProcessedLog {
    
    public final String filename;    
    public Forest forest;
    public ReconstructedForest recForest;
    public boolean overflowed;
    
    public ProcessedLog(String filename) {      
      this.filename=filename;
      this.forest=null;
      this.recForest=null;
      this.overflowed=false;
    }
    
    private final void processSKBLPPHeader(BufferedReader br, Forest f) {      
      try {
        String s1, s2;
        StringTokenizer st;      
      
        // grab blppHashEvents and kblppHashEvents
        s1 = br.readLine();
        if (s1==null) brokenLogfile("missing information on HashMapRVM accesses");
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile("HashMapRVM accesses info row is broken");
        s2 = st.nextToken();      
        // c <kbllpHashEvents> <blppHashEvents>
        if (!s2.equals("c") || !st.hasMoreTokens()) brokenLogfile("HashMapRVM accesses info row is broken");
        forest.kblppHashEvents=parseLong(st.nextToken());
        if (!st.hasMoreTokens()) brokenLogfile("HashMapRVM accesses info row is broken");
        forest.blppHashEvents=parseLong(st.nextToken());

        if (st.hasMoreTokens()) brokenLogfile("unknown field(s) in HashMapRVM accesses info row");
                  
        if (!br.markSupported()) { // will need to go back in the stream!
          System.out.println("Unknown error - exiting...");
          System.exit(1);
        }
                  
        br.mark(256); // arbitrary large value                
        boolean headerEnded = false;
        while ((s1 = br.readLine())!=null) {
          ProcessLog.currentRow++;
          st = new StringTokenizer(s1, " ");
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          s2 = st.nextToken();
          if (s2.equals("v") || s2.equals("p")) {
            headerEnded = true;
            br.reset(); // go backwards
            ProcessLog.currentRow--;
            break;  
          }          
          if (s2.equals("c") && !st.hasMoreTokens()) brokenLogfile("a routine info row is broken"); // TODO controllare
          
          RoutineStats stats = new RoutineStats();
          int cmid;
          
          cmid=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.avgIterations=parseFloat(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.maxIterations=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.minIterations=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.stdDeviation=parseFloat(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken"); 
          
          stats.totallyProcessed=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.numReset=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
                    
          stats.numBLPPevents=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a routine info row is broken");
          
          stats.numSpecialExits=parseInt(st.nextToken());
          
          if (st.hasMoreTokens()) { // TODO per vecchi log lo teniamo ancora                        
            if (parseInt(st.nextToken())>0) {
              stats.hasOverflowed=true;
              overflowed=true;
              System.out.println("WARNING: overflow in routine "+cmid);    
            }  
          }
          
          if (st.hasMoreTokens()) brokenLogfile("unknown field(s) in a routine info row");
          
          Routine r = new Routine(cmid);
          if (ProcessLog.reconstructRoutineGraph) r.routineGraph=getRoutineGraph(cmid);
          forest.routines.put(Integer.valueOf(cmid), r);
          r.routineStats = stats;          
          forest.numRoutines++;
          
          br.mark(256);
        }
        if (!headerEnded) brokenLogfile("premature end of file");        
      } catch (Exception e) {
        System.out.println("Unknown error - exiting...");
        System.out.println(e);        
        System.exit(1);
      }        
    }    
    
    private final void updateHashmaps(Node root) {      
      if (!forest.forestTreeRoots.add(root))
        brokenLogfile("root node ("+root.cmid+", "+root.pathNum+") was already in the set");
      
      Integer key = Integer.valueOf(root.cmid);
      Routine r = forest.routines.get(key);
      if (r==null) {
        if (forest.tool==Tool.SKBLPP) brokenLogfile("missing information for routine "+root.cmid);
        r = new Routine(root.cmid);
        r.routineStats=new RoutineStats();
        if (ProcessLog.reconstructRoutineGraph) r.routineGraph=getRoutineGraph(root.cmid);    
        forest.routines.put(key, r);
        forest.numRoutines++;
      }
      r.treeRoots.add(root);
      if (root.type!=NodeType.BLPP) r.containsCycles=true;
    }    
    
    public final void process() {
      try {
        // open file as a stream
        FileInputStream fstream = new FileInputStream(filename);
        System.out.println("Processing file "+filename+"...");
        DataInputStream in = new DataInputStream(fstream);
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        
        // useful for debug
        ProcessLog.currentFilename=filename;
        ProcessLog.currentRow=0;
        
        String s1, s2;        
        StringTokenizer st;
                
        // process first row of the log to determine which tool produced id
        s1 = br.readLine();
        if (s1==null) brokenLogfile("empty log file");
        ProcessLog.currentRow++;
        
        st = new StringTokenizer(s1, " ");
        if (!st.hasMoreTokens()) brokenLogfile("first row is broken");
        s2 = st.nextToken();
        if (!s2.equals("c") || !st.hasMoreTokens()) brokenLogfile("first row is broken");
        s2 = st.nextToken();
        
        Tool tool = null;
        if (s2.equals("blpp")) tool=Tool.BLPP;
        else if (s2.equals("kblpp")) tool=Tool.KBLPP;
        else if (s2.equals("tblpp")) tool=Tool.TBLPP;
        else if (s2.equals("tkblpp")) tool=Tool.TKBLPP;
        else if (s2.equals("skblpp")) tool=Tool.SKBLPP;
        else brokenLogfile("unknown tool");
        
        int bufsize=0;
        int chosenK=0;
               
        if (tool==Tool.TKBLPP || tool==Tool.SKBLPP) { // retrieve chosenK and bufferSize
          if (!st.hasMoreTokens()) brokenLogfile("first row is broken");          
          chosenK=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("first row is broken");
          bufsize=parseInt(st.nextToken());
        } else if (tool==Tool.TBLPP) { // retrieve bufferSize
          if (!st.hasMoreTokens()) brokenLogfile("first row is broken");
          bufsize=parseInt(st.nextToken());
        } else if (tool==Tool.KBLPP) {
          if (!st.hasMoreTokens()) brokenLogfile("first row is broken");          
          chosenK=parseInt(st.nextToken());  
        }    
        
        // else: nothing to retrieve for BLPP        
        if (st.hasMoreTokens()) brokenLogfile("unknown field(s) in first row");  
        
        System.out.println("Tool: "+tool);
        if (lastTool==null)
          lastTool=tool;
        else if (lastTool!=tool) {
          System.out.println("Error: processing log files produced by different profiling tools");
          System.exit(1);
        }                
        
        if (chosenK>0) {
          System.out.println("Chosen K: "+chosenK);
          if (lastK==-1)
            lastK=chosenK;
          else if (lastK != chosenK) {
            System.out.println("Error: processing log files produced with different values for k");
            System.exit(1);              
          }             
        }    
    
        if (bufsize>0)
          System.out.println("Buffer size: "+bufsize); // TODO come sopra
                
        
        // initialize Forest
        forest=new Forest(tool, chosenK, bufsize);        
        if (tool==Tool.SKBLPP)
          processSKBLPPHeader(br, forest);          
        
        boolean nodesEnded = false;
        Stack<Node> stack = new Stack<Node>();
                
        while ((s1 = br.readLine())!=null) {
          ProcessLog.currentRow++;
          st = new StringTokenizer(s1, " ");
          if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
          s2 = st.nextToken();
          if (s2.equals("p")) {
            nodesEnded = true;            
            break;  
          }
        
          if (!s2.equals("v")) brokenLogfile("unexpected row in file");                    
          if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");

          int nodeID, parentID, cmid, pathNum, count;
          NodeType type;
          boolean isSpecial=false;
          
          if (tool!=Tool.BLPP && tool!=Tool.TBLPP) {
            nodeID=parseInt(st.nextToken(), 16);
            if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
          
            parentID=parseInt(st.nextToken(), 16);          
            if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
          } else {
            nodeID=0;
            parentID=0; 
          }  
          
          cmid=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
          
          pathNum=parseInt(st.nextToken());
          if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
          
          count=parseInt(st.nextToken());          
          if (count<0) {
            overflowed=true;
            System.out.println("WARNING: overflow detected in "+ProcessLog.currentFilename+":"+ProcessLog.currentRow);
          }  
          
          if (tool!=Tool.BLPP && tool!=Tool.TBLPP) {
            if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");
            int typeInt=parseInt(st.nextToken());
            if (!st.hasMoreTokens()) brokenLogfile("a node row is broken");          
            type = NodeType.getTypeForInt(typeInt);
            if (type==null) brokenLogfile("invalid node type");
          
            if (parseInt(st.nextToken())>0) isSpecial=true;
          } else {
            type = NodeType.BLPP;
            isSpecial = false;
          }    
          
          if (st.hasMoreTokens()) brokenLogfile("unexpected field(s) in a node row");
          
          // process row          
          if (parentID==0) {
            Node node = new Node(nodeID, cmid, pathNum, type, isSpecial, count, null);
            stack.removeAllElements();
            if (type!=NodeType.BLPP) stack.push(node);
            updateHashmaps(node);
          } else {
            Node parent = null;
            while (!stack.isEmpty()) {
              parent = stack.pop();
              if (parent.nodeID==parentID) break;  
            }
            if (parent==null || parent.nodeID!=parentID) {
              brokenLogfile("graph is not a forest");
            } else {
              stack.push(parent);
              Node node = new Node(nodeID, cmid, pathNum, type, isSpecial, count, parent);
              if (parent.firstChild==null) parent.firstChild=node;
              else {
                Node tmp;
                for (tmp=parent.firstChild; tmp.nextSibling!=null; tmp=tmp.nextSibling);
                tmp.nextSibling=node;  
              }              
              stack.push(node);  
            }
          }  
          forest.numNodes++;                
        }
        
        if (!nodesEnded) brokenLogfile("premature end of file");
        
        if (!st.hasMoreTokens()) brokenLogfile("forest info row is broken");
        
        int numNodes = parseInt(st.nextToken());
        if (numNodes!=forest.numNodes) brokenLogfile("wrong number of nodes");        
        
        if (forest.tool==Tool.BLPP || forest.tool==Tool.TBLPP)
          System.out.println("Nodes: "+numNodes);
        else
          System.out.println("k-SF nodes: "+numNodes);  
        
        System.out.println("Hot routines: "+forest.numRoutines);
        
        if (st.hasMoreTokens()) brokenLogfile("unexpected field(s) in forest info row");

        // close file and reset global fields
        ProcessLog.currentFilename=null;
        ProcessLog.currentRow=-1;        
        in.close();
        fstream.close();      
      } catch (FileNotFoundException e1) {
        System.out.println("Unable to open file \""+filename+"\" - exiting...");
        System.out.println(e1);
        System.exit(1);  
      } catch (IOException e2) {
        System.out.println("Unknown error when processing file \""+filename+"\" - exiting...");
        System.out.println(e2);
        System.exit(1);
      }
      
      // reconstruct number of KBLPP hash accesses (not needed for SKBLPP)
      if (forest.tool==Tool.KBLPP || forest.tool==Tool.TKBLPP) {
        for (Iterator<Node> itr = forest.forestTreeRoots.iterator(); itr.hasNext(); ) {
          Node root = itr.next();
          if (root.type!=NodeType.BLPP)            
            forest.kblppHashEvents+=1+root.count;
        }     
      }
    }
    
    public final void cleanCounters() {
      for (Iterator<Node> itr = forest.forestTreeRoots.iterator(); itr.hasNext(); ) {
        Node root = itr.next();
        if (root.type==NodeType.BLPP) continue;        
        if (root.type!=NodeType.BEGIN)
          Routine.cleanCounters(root, 0, forest.chosenK);  
      }    
    }  
    
    public final void performReconstruction(int chosenK) {      
      recForest = new ReconstructedForest(forest);
      recForest.performReconstruction(chosenK);      
    }
    
    public final void performReconstruction() {
      performReconstruction(forest.chosenK);  
    }
    
    public final void performCounterAnalysis(ProcessedLogStats stats) {
      forest.performCounterAnalysis(stats);    
    }    
    
    public final void dumpKSLAB(boolean generateSVG) {
      forest.dumpKSLAB(filename, generateSVG);    
    }
    
    public final void dumpReconstructedForest(boolean generateSVG) {
      recForest.dumpReconstructedForest(filename, generateSVG);  
    }    
    
    public final void analyzeIPF(ProcessedLogStats stats) {
      recForest.analyzeIPF(stats);
    }
    
    public final void clean() {
      if (forest!=null) forest.clean();
      if (recForest!=null) recForest.clean();  
    }            
  }
        
  private static boolean inputYesNo(String question, Scanner scanner) {
    System.out.print(question);
    String answer = scanner.nextLine();
    if (answer.equals("y") || answer.equals("Y"))
        return true;
    else if (!answer.equals("n") && !answer.equals("N")) {
        System.out.println("Invalid choice. Exiting...");
        System.exit(1);
    }
    return false;
  }
  
  private static float inputThreshold(String question, Scanner scanner, float min, float max) {
    System.out.print(question);
    if (!scanner.hasNextFloat()) {    
        System.out.println("Invalid choice. Exiting...");
        System.exit(1);
    }
    float f = scanner.nextFloat();    
    if (f<min || f>max) {
        System.out.println("Value out of range. Exiting...");
        System.exit(1);
    }
    return f;
  }    
          
        
  public static void main(String[] args) {
    if (args.length==0) {
      // process only one logfile (for now)
      System.out.println("Usage: java ProcessLog logfile.log [<logfile2.log> <...>]");
      System.exit(1);
    }
    
    // ask the user for some parameters
    Scanner scanner = new Scanner(System.in);
    ProcessLog.dumpKSFs = inputYesNo("Should I produce a graphical representation of the k-SFs? [y/n]: ", scanner);    
    ProcessLog.dumpKIPFs = inputYesNo("Should I produce a graphical representation of the k-IPFs? [y/n]: ", scanner);
    if (dumpKIPFs) {        
      ProcessLog.pruneKIPFs = inputYesNo("Shoud I prune the k-IPFs? [y/n]: ", scanner);    
      if (pruneKIPFs) {
        ProcessLog.rootThreshold = inputThreshold("Minimal hotness threshold for a tree root wrt routine counters [0;1]: ", scanner, 0, 1);      
        ProcessLog.treeRootThreshold = inputThreshold("Minimal hotness threshold for a child node wrt its root [0;1]: ", scanner, 0, 1);
        ProcessLog.childThreshold = inputThreshold("Minimal hotness threshold for a child node wrt its parent [0;1]: ", scanner, 0, 1);
      }
    }  
        
    // check whether routine/ exists
    File f1 = new File(args[0]);
    ProcessLog.routineDirName=f1.getParent();
    if (routineDirName==null)
      routineDirName="routine";
    else
      routineDirName+="/routine";
    File f2 = new File(routineDirName);
    if (f2.exists() && f2.isDirectory()) reconstructRoutineGraph = true;
    
    ProcessedLog[] logs = new ProcessedLog[args.length];
    ProcessedLogStats[] stats = new ProcessedLogStats[args.length];
    
    ProcessedLogStats globalStats=new ProcessedLogStats();
    boolean overflowDetected = false;
    
    for (int i=0; i<args.length; i++) {
      ProcessedLog pf = new ProcessedLog(args[i]);
      logs[i] = pf;      
      stats[i] = new ProcessedLogStats();
      pf.process();
      overflowDetected = overflowDetected || pf.overflowed;
      pf.cleanCounters();
      pf.performCounterAnalysis(stats[i]);      
      if (ProcessLog.dumpKSFs)
        pf.dumpKSLAB(true);
      pf.performReconstruction();
      pf.analyzeIPF(stats[i]);
      if (ProcessLog.dumpKIPFs)
        pf.dumpReconstructedForest(true);
      ProcessedLogStats.updateProgramStats(globalStats, stats[i]);
      pf.clean(); // gargage collection is actually needed for large files :)
      System.out.println();
    }
    
    // dump cumulative statistics for all threads
    DecimalFormat myFormat = new DecimalFormat("#0.####");
    String s; double fraction;
    
    System.out.println("=========================");
    System.out.println("Program global statistics");
    System.out.println("=========================");
    System.out.println("Profiling tool: "+lastTool);
    int spannedIterations=1;
    if (lastK > 0) {
      System.out.println("Chosen K: "+lastK);
      spannedIterations += lastK;
    }      
    System.out.println("Spanned iterations: "+spannedIterations);  
    int numRoutines = graphMap.size();
    System.out.println("Hot routines: "+numRoutines);
    
    System.out.println("Overflow detected: "+overflowDetected);    
    if (lastTool == Tool.BLPP || lastTool == Tool.TBLPP) {
      System.out.println("Nodes: "+globalStats.numBLPPNodes);
      System.out.println("Processed events: "+globalStats.processedEvents);
      System.out.println("Hash get events: "+globalStats.numBLPPGetEvents);    
      System.out.println("Hash put events: "+globalStats.numBLPPPutEvents);
      System.out.println("Total hash accesses: "+(globalStats.numBLPPGetEvents+globalStats.numBLPPPutEvents));    
      System.out.println("Sum of BLPP counters: "+globalStats.sumOfBLPPcounters);
    } else {
      System.out.println("k-SF nodes: "+(globalStats.numBLPPNodes+globalStats.numKBLPPNodes));
      System.out.println("k-SF nodes for out-of-cycle paths: "+globalStats.numBLPPNodes);
      System.out.println("k-SF nodes for k-iteration paths: "+globalStats.numKBLPPNodes);
      System.out.println("k-SF trees for k-iteration paths: "+globalStats.numKBLPPTrees);    
      System.out.println("Number of k-SF internal nodes: "+globalStats.numKSFinternalNodes);
      s = myFormat.format(globalStats.avgKSFdegree);
      System.out.println("Avg k-SF node degree: "+s);    
      s = myFormat.format(globalStats.avgKSFperChildLookup);
      System.out.println("Avg k-SF child lookups: "+s);
      System.out.println("Processed events: "+globalStats.processedEvents);
      System.out.println("Hash get events for out-of-cycle paths: "+globalStats.numBLPPGetEvents);    
      System.out.println("Hash put events for out-of-cycle paths: "+globalStats.numBLPPPutEvents);
      System.out.println("Total hash accesses for out-of-cycle paths: "+(globalStats.numBLPPGetEvents+globalStats.numBLPPPutEvents));    
      System.out.println("Hash get events for k-iteration paths: "+globalStats.numKBLPPGetEvents);    
      System.out.println("Hash put events for k-iteration paths: : "+globalStats.numKBLPPPutEvents);    
      System.out.println("Total hash accesses for k-iteration paths: "+(globalStats.numKBLPPGetEvents+globalStats.numKBLPPPutEvents));
      System.out.println("Total hash accesses: "+(globalStats.numBLPPGetEvents+globalStats.numBLPPPutEvents
                                                +globalStats.numKBLPPGetEvents+globalStats.numKBLPPPutEvents));
      System.out.println("Sum of k-SF counters for out-of-cycle paths: "+globalStats.sumOfBLPPcounters);
      System.out.println("Sum of k-SF counters for k-iteration paths: "+globalStats.sumOfKBLPPcounters);
    
    
      fraction = globalStats.computeFractionOfAccessesForCyclesWrtBLPP();
      s=myFormat.format(fraction);
      System.out.println("Fraction of hash accesses made for k-iteration paths wrt traditional BLPP: "+s);
      
      fraction = globalStats.computeFractionOfAccessesWrtBLPP();
      s=myFormat.format(fraction);
      System.out.println("Fraction of total hash accesses wrt traditional BLPP: "+s);        
    }      
    
    System.out.println("Number of k-IPF nodes: "+globalStats.numIPFnodes);    
    System.out.println("Number of k-IPF roots: "+globalStats.numIPFroots);
    System.out.println("Number of k-IPF leaves: "+globalStats.numIPFleaves);
    System.out.println("Sum of k-IPF root counters: "+globalStats.sumIPFrootCounters);    
    
    s = myFormat.format(globalStats.avgIPFleafDepth);
    System.out.println("Avg k-IPF leaf depth: "+s);
    
    System.out.println("Number of k-IPF internal nodes: "+globalStats.numIPFinternalNodes);
    s = myFormat.format(globalStats.avgIPFdegree);
    System.out.println("Avg k-IPF node degree: "+s);
  
    if (lastTool == Tool.SKBLPP) {
      for (int i=0; i<globalStats.valuesForK.length; i++) {
        if (globalStats.valuesForK[i]>lastK) break;
        if (globalStats.sumOfKBLPPcounters>0)
          fraction = ((double) globalStats.eventsWithinK[i])/(globalStats.sumOfKBLPPcounters);
        else fraction=0;  
        s = myFormat.format(fraction);
        System.out.println("Fraction of KBLPP events inside loops of length within "+(globalStats.valuesForK[i]+1)+
                           " (k="+globalStats.valuesForK[i]+"): "+s);
      }
    }
    
    String header = "#iterations numHotRoutines processedEvents acyclicKSFnodes cyclicKSFnodes totalKSFNodes acyclicHashes cyclicHashes totalHashes "+
                    "hashFractionInCyclesOnly hashFractionGlobal sumAcyclicKSFcounters sumCyclicKSFcounters numIPFnodes numIPFtrees sumIPFrootCounters "+
                    "numIPFleaves avgIPFleafDepth numIPFinternal avgIPFdegree numKSFinternal avgKSFdegree avgKSFchildLookups";
    String dataRow = spannedIterations+" "+numRoutines+" "+globalStats.toString();
    
    System.err.println(header);
    System.err.println(dataRow);
                        
    System.exit(0);
  }

}
