import java.io.*;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Iterator;

public final class ReconstructedForest {
  public final Forest forest;
  public int usedK;
  public ArrayList<ReconstructedRoutine> routines;
  
  public ReconstructedForest(Forest forest) {
    this.forest=forest;  
    this.usedK=-1;
    this.routines=new ArrayList<ReconstructedRoutine>();
  }
  
  public final void clean() {
    for (Iterator<ReconstructedRoutine> itr=routines.iterator(); itr.hasNext(); ) {
      ReconstructedRoutine rec = itr.next();
      rec.clean();
    }
    routines.clear();
    routines=null;  
  }    
  
  public final void performReconstruction(int chosenK) {
    // set usedK
    if (chosenK < 0 || chosenK > forest.chosenK) {
      System.out.println("WARNING: chosen K is not valid - using maximum allowed value ("+forest.chosenK+") instead!");  
      usedK=forest.chosenK;  
    } else usedK=chosenK;
    // perform reconstruction
    for (Routine r : forest.routines.values()) {
      //~ if (!r.containsCycles) continue;
      ReconstructedRoutine rec = r.prepareForReconstruction(forest);
      rec.performReconstruction(usedK);
      routines.add(rec);
    }    
  }
  
  public final void dumpReconstructedForest(String filename, boolean generateSVG) {
    String threadName;
    if (filename.length()<=4) {
      System.out.println("Please do not rename logfiles!");
      threadName = filename+"-out";
    } else {   
      threadName = filename.substring(0, filename.length()-4);
    }    
    String destDir = threadName+"/reconstructed-"+usedK;        
    System.out.println("Creating output files in "+destDir+"/ folder...");
    
    File dir = new File(destDir+"/dot");
    dir.mkdirs();
    
    Runtime rt = null;    
    if (generateSVG) {
      dir = new File(destDir+"/svg");
      dir.mkdirs();
      rt = Runtime.getRuntime();
    }  
    
    for (Iterator<ReconstructedRoutine> itr=routines.iterator(); itr.hasNext(); ) {
      ReconstructedRoutine rec = itr.next();
      if (!rec.routine.containsCycles) continue;
      String dotname = destDir + "/dot/" + rec.routine.cmid + ".dot";
      PrintStream ps = null;
      try {
        ps = new PrintStream(new FileOutputStream(dotname));
      } catch (IOException e) {
        System.out.println("Error: cannot open output file "+dotname+ " - exiting...");
        System.out.println(e);
        System.exit(1);
      }    
      rec.dumpReconstructed(ps, this);
      ps.close();
        
      if (generateSVG) {
        String svgname = destDir + "/svg/" + rec.routine.cmid + ".svg";
        String command = "/usr/bin/dot -Tsvg "+dotname+" -o "+svgname;      
        try {      
          Process pr = rt.exec(command);
          //~ pr.waitfor();
        } catch (IOException e) {
          System.out.println("Warning: unable to generate SVG graph "+svgname);
          System.out.println(e);  
        }        
      }
    }  
  }
  
  public final void analyzeIPF(ProcessedLogStats stats) {  
    for (Iterator<ReconstructedRoutine> itr=routines.iterator(); itr.hasNext(); ) {
      ReconstructedRoutine rec = itr.next();
      // k-BLPP nodes
      for (Iterator<Node> routineItr=rec.reconstructedTrees.iterator(); routineItr.hasNext(); ) {    
        Node root = routineItr.next();
        analyzeIPFaux(root, 0, stats);
      }
      // BLPP nodes
      for (Iterator<Node> routineItrBLPP=rec.blppNodes.iterator(); routineItrBLPP.hasNext(); ) {          
        Node blppRoot = routineItrBLPP.next();
        analyzeIPFaux(blppRoot, 0, stats);
      }
    }          
    if (stats.numIPFleaves>0)
      stats.avgIPFleafDepth=((double)stats.sumIPFleafDepths)/stats.numIPFleaves;      
    if (stats.numIPFinternalNodes>0)
      stats.avgIPFdegree=((double)stats.numIPFchildren)/stats.numIPFinternalNodes;
    
    DecimalFormat myFormat = new DecimalFormat("#0.####");
    String s;
      
    System.out.println("Number of k-IPF nodes: "+stats.numIPFnodes);
    System.out.println("Number of k-IPF roots: "+stats.numIPFroots);
    System.out.println("Number of k-IPF leaves: "+stats.numIPFleaves);
    System.out.println("Sum of k-IPF root counters: "+stats.sumIPFrootCounters);    
    
    s = myFormat.format(stats.avgIPFleafDepth);
    System.out.println("Avg k-IPF leaf depth: "+s); 
    
    System.out.println("Number of k-IPF internal nodes: "+stats.numIPFinternalNodes);
    s = myFormat.format(stats.avgIPFdegree);
    System.out.println("Avg k-IPF node degree: "+s);
  }
    
  private final void analyzeIPFaux(Node node, int depth, ProcessedLogStats stats) {
    if (node==null) return;
    
    stats.numIPFnodes++;
    
    if (node.parent==null) {
      stats.numIPFroots++;
      stats.sumIPFrootCounters+=node.count;
    }
           
    if (node.firstChild==null) {
      stats.numIPFleaves++;
      stats.sumIPFleafDepths+=depth;
    } else {       
      stats.numIPFinternalNodes++;
      for (Node child = node.firstChild; child != null; child=child.nextSibling) {
        analyzeIPFaux(child, depth + 1, stats);
        stats.numIPFchildren++;
      }  
    }    
  }    
}
