import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public final class ReconstructedRoutine {  
  public Routine routine;
  public ArrayList<Node> reconstructedTrees;
  public ArrayList<Node> blppNodes;
  public HashMap<Integer, ArrayList<Node>> labels;
  
  public ReconstructedRoutine(Routine routine) {
    this.routine=routine;
    this.reconstructedTrees=new ArrayList<Node>();
    this.blppNodes=new ArrayList<Node>();
    this.labels=new HashMap<Integer, ArrayList<Node>>();
  }
  
  public final void clean() {
    for (Iterator<Node> itr=reconstructedTrees.iterator(); itr.hasNext(); ) {    
      Node root = itr.next();
      Node.cleanTree(root);
    }
    reconstructedTrees.clear();
    reconstructedTrees=null;
    
    for (Iterator<Node> itr=blppNodes.iterator(); itr.hasNext(); ) {    
      Node root = itr.next();
      Node.cleanTree(root);
    }
    blppNodes.clear();
    blppNodes=null;
    
    // nodes inside labels must not be freed here!
    for (ArrayList<Node> list: labels.values())
      list.clear();
    labels.clear();
    labels=null;
  }    
  
  public final void performReconstruction(int chosenK) {
    // process labels and populate reconstructedTrees
    for (Map.Entry<Integer, ArrayList<Node>> iterator : labels.entrySet()) {
      Integer pathNum = iterator.getKey();
      ArrayList<Node> nodes = iterator.getValue();
           
      // labels contain BEGIN, NORMAL and EXIT nodes: separate them (at first level only)
      @SuppressWarnings("unchecked")
      ArrayList<Node>[] roots = new ArrayList[3];
      roots[0]=new ArrayList<Node>();
      roots[1]=new ArrayList<Node>();
      roots[2]=new ArrayList<Node>();
      
      for (Iterator<Node> itr = nodes.iterator(); itr.hasNext(); ) {
        Node node = itr.next();
        switch (node.type) {
          case BEGIN: roots[0].add(node); break;
          case NORMAL: roots[1].add(node); break;
          case EXIT: roots[2].add(node); break;
          default: break;
        }
      }
      
      for (int i=0; i<3; i++) {
        if (roots[i].isEmpty()) continue;
        NodeType type = roots[i].get(0).type;
        Node root = new Node(0, routine.cmid, pathNum, type, false, 0, null);
        for (Iterator<Node> itr = roots[i].iterator(); itr.hasNext(); ) {
          Node node = itr.next();
          joinTrees(root, node, chosenK);
        }    
        reconstructedTrees.add(root);
      }              
    }
  }
  
  public final void dumpReconstructed(PrintStream ps, ReconstructedForest rec) {    
    ps.println("digraph {");
    int tmp=0;
        
    long sumOfRootCounters=0;
    if (ProcessLog.pruneKIPFs) {
      for (Iterator<Node> itr=reconstructedTrees.iterator(); itr.hasNext(); ) {    
        Node root = itr.next();
        sumOfRootCounters+=root.count;    
      }
    }  
        
    // print reconstructed trees  
    if (ProcessLog.pruneKIPFs) {
      for (Iterator<Node> itr=reconstructedTrees.iterator(); itr.hasNext(); ) {    
        Node root = itr.next();      
        if (root.count<sumOfRootCounters*ProcessLog.rootThreshold) continue; // first pruning stage    
        ps.println("subgraph s"+(tmp++)+" {");        
        ProcessLog.dumpGraphvizAuxWithPruning(null, root, ps, root); // second pruning stage
        ps.println("}");  
      }
    } else {
      for (Iterator<Node> itr=reconstructedTrees.iterator(); itr.hasNext(); ) {      
        Node root = itr.next();
        ps.println("subgraph s"+(tmp++)+" {");
        ProcessLog.dumpGraphvizAux(null, root, ps);      
        ps.println("}");        
      }    
    }        
    
    // then print BLPP nodes in a table
    if (!blppNodes.isEmpty()) {
      ps.println("subgraph s"+(tmp++)+" {");
      ProcessLog.dumpGraphvizAuxBLPPAsTable(blppNodes, ps);
      ps.println("}");
    }
       
    // print some statistics
    ps.println("subgraph s"+(tmp++)+"{");
    ps.println("Statistics [shape=none, margin=0, label=<");
    ps.println("<TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"2px\" CELLPADDING=\"2px\">");
    ps.println("<TH><TD>Statistics</TD></TH>");
    ps.println("<TR><TD>Events</TD><TD>"+sumOfRootCounters+"</TD></TR>");
    
    
    if (ProcessLog.rootThreshold > 0 || ProcessLog.treeRootThreshold > 0 || ProcessLog.childThreshold > 0) {
      ps.println("<TR><TD>Alpha</TD><TD>"+String.format("%f", ProcessLog.rootThreshold)+"</TD></TR>");    
      ps.println("<TR><TD>Beta</TD><TD>"+String.format("%f", ProcessLog.treeRootThreshold)+"</TD></TR>");
      ps.println("<TR><TD>Gamma</TD><TD>"+String.format("%f", ProcessLog.childThreshold)+"</TD></TR>");
    }    
  
    if (ProcessLog.reconstructRoutineGraph) {
      ps.println("<TR><TD>Chosen K</TD><TD>"+(rec.forest.chosenK+1)+"</TD></TR>");
      ps.println("<TR><TD>Routine ID</TD><TD>"+routine.cmid+"</TD></TR>");            
      ps.println("<TR><TD>Method</TD><TD>"+routine.routineGraph.methodName+"</TD></TR>");
      ps.println("<TR><TD>Class</TD><TD>"+routine.routineGraph.shortClassName+"</TD></TR>");
      ps.println("<TR><TD>Package</TD><TD>"+routine.routineGraph.pkgName+"</TD></TR>");
      ps.println("<TR><TD>Src file</TD><TD>"+routine.routineGraph.srcFile+"</TD></TR>");
      ps.println("<TR><TD>Opt level</TD><TD>"+routine.routineGraph.optLevel+"</TD></TR>");
    }  
  
    // detailed statistics (SKBLPP only)
    if (rec.forest.tool==Tool.SKBLPP) {
      ps.println("<TR><TD>Avg iterations</TD><TD>"+routine.routineStats.avgIterations+"</TD></TR>");
      ps.println("<TR><TD>Std deviation</TD><TD>"+routine.routineStats.stdDeviation+"</TD></TR>");    
      ps.println("<TR><TD>Min iterations</TD><TD>"+routine.routineStats.minIterations+"</TD></TR>");
      ps.println("<TR><TD>Max iterations</TD><TD>"+routine.routineStats.maxIterations+"</TD></TR>");
        
      ps.println("<TR><TD>k-BLPP events</TD><TD>"+routine.routineStats.totallyProcessed+"</TD></TR>");
      ps.println("<TR><TD>Reset events</TD><TD>"+routine.routineStats.numReset+"</TD></TR>");
      ps.println("<TR><TD>BLPP events</TD><TD>"+routine.routineStats.numBLPPevents+"</TD></TR>");
      ps.println("<TR><TD>Special exits</TD><TD>"+routine.routineStats.numSpecialExits+"</TD></TR>");
    }      
    
    ps.println("</TABLE>");
    ps.println(">];");
    ps.println("}");
    
    // end of graph
    ps.println("}");   
  }  
  
  private static final void joinTrees(Node first, Node second, int depth) {
    if (first==null || second==null) return;
    
    first.count+=second.count;
    first.isSpecial=first.isSpecial || second.isSpecial;
    
    // TODO might never happen      
    if (first.type!=second.type)
      System.out.println("Warning: distinct types ("+first.type+", "+second.type+") for same node label ("+first.pathNum+") in routine "+first.cmid);
    
    if (depth<=0 || second.firstChild==null) return;
        
    for (Node childOfSecond=second.firstChild; childOfSecond!=null; childOfSecond=childOfSecond.nextSibling) {
      Node childOfFirst;
      for (childOfFirst=first.firstChild; childOfFirst!=null; childOfFirst=childOfFirst.nextSibling)
        if (childOfFirst.pathNum==childOfSecond.pathNum && childOfFirst.type==childOfSecond.type) break;
      if (childOfFirst==null) {
        // count will be populated by joinTrees
        childOfFirst=new Node(0, childOfSecond.cmid, childOfSecond.pathNum, childOfSecond.type, childOfSecond.isSpecial, 0, first);
        // TODO ordered insertion by count ???
        if (first.firstChild==null) first.firstChild=childOfFirst;
        else {
          Node tmp;
          for (tmp=first.firstChild; tmp.nextSibling!=null; tmp=tmp.nextSibling);
          tmp.nextSibling=childOfFirst;  
        }    
      }
      joinTrees(childOfFirst, childOfSecond, depth-1);                
    }    
  }
}    
 
