import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

public final class Routine {
  public int cmid;
  public RoutineGraph routineGraph; // TODO
  public RoutineStats routineStats;
  public ArrayList<Node> treeRoots;
  public boolean containsCycles;
  
  public Routine(int cmid) {
    this.cmid=cmid;
    this.routineGraph=null;
    this.routineStats=null;
    this.treeRoots=new ArrayList<Node>();  
    this.containsCycles=false;
  }
  
  public final void clean() {
    routineGraph=null; // TODO
    routineStats=null;        
    // clean nodes recursively
    for (Iterator<Node> itr = treeRoots.iterator(); itr.hasNext(); ) {
      Node root = itr.next();
      Node.cleanTree(root);
    }  
    treeRoots.clear();
    treeRoots=null;  
  }
  
  public final void performCounterAnalysis(ProcessedLogStats stats) {
    for (Iterator<Node> itr = treeRoots.iterator(); itr.hasNext(); ) {
      Node root = itr.next();      
      if (root.type==NodeType.BLPP) {
        routineStats.sumOfBLPPcounters+=root.count;
        stats.numBLPPNodes++;
      } else {
        stats.numKBLPPTrees++;
        routineStats.sumOfKBLPPcounters+=sumTreeCounters(root, stats, 0); // position is zero for tree roots
      }  
    }
  }
  
  // performs also avg degree analysis for the k-SF
  private final static long sumTreeCounters(Node node, ProcessedLogStats stats, int position) {
    if (node==null) return 0;
    
    stats.numKBLPPNodes++;  
    long ret = node.count;
    
    if (node.parent!=null) {
      stats.sumKSFchildLookups+=position*node.count;
      stats.sumKSFchildCounters+=node.count;
    }   
    
    if (node.firstChild!=null) stats.numKSFinternalNodes++;
    
    int p=1;
    for (Node child=node.firstChild; child!=null; child=child.nextSibling) {      
      ret += sumTreeCounters(child, stats, p);
      stats.numKSFchildren++;
      p++;
    }  
    
    return ret;  
  }    
  
  public final void dumpKSLAB(PrintStream ps, Forest forest) {
    ps.println("digraph {");
    int tmp=0;
    
    ArrayList<Node> blppNodes = new ArrayList<Node>();
    
    for (Iterator<Node> itr = treeRoots.iterator(); itr.hasNext(); ) {
      Node root = itr.next();
      if (root.type==NodeType.BLPP) {
        blppNodes.add(root);
        continue;
      }      
      ps.println("subgraph s"+(tmp++)+" {");
      ProcessLog.dumpGraphvizAux(null, root, ps);
      ps.println("}");
    }
    
    // dump blppNodes in a table
    if (!blppNodes.isEmpty()) {
      ps.println("subgraph s"+(tmp++)+" {");
      ProcessLog.dumpGraphvizAuxBLPPAsTable(blppNodes, ps);
      ps.println("}");
    }    
          
    // print statistics
    ps.println("subgraph s"+(tmp++)+"{");
    ps.println("Statistics [shape=none, margin=0, label=<");
    ps.println("<TABLE BORDER=\"0\" CELLBORDER=\"1\" CELLSPACING=\"2px\" CELLPADDING=\"2px\">");
    ps.println("<TH><TD>Statistics</TD></TH>");
  
    if (ProcessLog.reconstructRoutineGraph) {
      ps.println("<TR><TD>Chosen K</TD><TD>"+(forest.chosenK+1)+"</TD></TR>");
      ps.println("<TR><TD>Routine ID</TD><TD>"+cmid+"</TD></TR>");            
      ps.println("<TR><TD>Method</TD><TD>"+routineGraph.methodName+"</TD></TR>");
      ps.println("<TR><TD>Class</TD><TD>"+routineGraph.shortClassName+"</TD></TR>");
      ps.println("<TR><TD>Package</TD><TD>"+routineGraph.pkgName+"</TD></TR>");
      ps.println("<TR><TD>Src file</TD><TD>"+routineGraph.srcFile+"</TD></TR>");
      ps.println("<TR><TD>Opt level</TD><TD>"+routineGraph.optLevel+"</TD></TR>");
    }  
  
    // detailed statistics (SKBLPP only)
    if (forest.tool==Tool.SKBLPP) {
      ps.println("<TR><TD>Avg iterations</TD><TD>"+routineStats.avgIterations+"</TD></TR>");
      ps.println("<TR><TD>Std deviation</TD><TD>"+routineStats.stdDeviation+"</TD></TR>");    
      ps.println("<TR><TD>Min iterations</TD><TD>"+routineStats.minIterations+"</TD></TR>");
      ps.println("<TR><TD>Max iterations</TD><TD>"+routineStats.maxIterations+"</TD></TR>");
        
      ps.println("<TR><TD>k-BLPP events</TD><TD>"+routineStats.totallyProcessed+"</TD></TR>");
      ps.println("<TR><TD>Reset events</TD><TD>"+routineStats.numReset+"</TD></TR>");
      ps.println("<TR><TD>BLPP events</TD><TD>"+routineStats.numBLPPevents+"</TD></TR>");
      ps.println("<TR><TD>Special exits</TD><TD>"+routineStats.numSpecialExits+"</TD></TR>");
    }      
    
    ps.println("</TABLE>");
    ps.println(">];");
    ps.println("}");
        
    // close digraph    
    ps.println("}");
  }
  
  // returns a <pathNum, list of nodes> HashMap
  public ReconstructedRoutine prepareForReconstruction(Forest forest) {
    ReconstructedRoutine rec = new ReconstructedRoutine(this);    
      
    for (Iterator<Node> itr = treeRoots.iterator(); itr.hasNext(); ) {
      Node root = itr.next();
      if (root.type==NodeType.BLPP) {
        rec.blppNodes.add(Node.cloneBLPPNode(root));
        continue;  
      }
      if (root.type!=NodeType.BEGIN) {      
        cleanCounters(root, 0, forest.chosenK);  
      }
      findNodesForLabels(root, rec.labels);               
    }
    
    return rec;
  }
  
  private static void findNodesForLabels(Node node, HashMap<Integer, ArrayList<Node>> map) {
    if (node==null) return;
    
    Integer pathNum=node.pathNum;
    ArrayList<Node> list = map.get(pathNum);
    if (list==null) {
      list = new ArrayList<Node>();  
      map.put(pathNum, list);
    }
    list.add(node);
    
    for (Node child=node.firstChild; child!=null; child=child.nextSibling)
      findNodesForLabels(child, map);
  }  
  
  // clean counters of nodes at depth [0, maxDepth-1]
  public static void cleanCounters(Node node, int depth, int maxDepth) {
    if (depth>=maxDepth || node == null) return;
    node.count=0;
    for (Node child=node.firstChild; child!=null; child=child.nextSibling)
    cleanCounters(child, depth+1, maxDepth);        
  }  

} 
